import { getFieldDef } from 'graphql/execution/execute';
import { isListType, isNonNullType, isObjectType } from 'graphql';
import { pathOr } from 'ramda';
import chalk from 'chalk';

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _inheritsLoose(subClass, superClass) {
  subClass.prototype = Object.create(superClass.prototype);
  subClass.prototype.constructor = subClass;
  subClass.__proto__ = superClass;
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function isNativeReflectConstruct() {
  if (typeof Reflect === "undefined" || !Reflect.construct) return false;
  if (Reflect.construct.sham) return false;
  if (typeof Proxy === "function") return true;

  try {
    Date.prototype.toString.call(Reflect.construct(Date, [], function () {}));
    return true;
  } catch (e) {
    return false;
  }
}

function _construct(Parent, args, Class) {
  if (isNativeReflectConstruct()) {
    _construct = Reflect.construct;
  } else {
    _construct = function _construct(Parent, args, Class) {
      var a = [null];
      a.push.apply(a, args);
      var Constructor = Function.bind.apply(Parent, a);
      var instance = new Constructor();
      if (Class) _setPrototypeOf(instance, Class.prototype);
      return instance;
    };
  }

  return _construct.apply(null, arguments);
}

function _isNativeFunction(fn) {
  return Function.toString.call(fn).indexOf("[native code]") !== -1;
}

function _wrapNativeSuper(Class) {
  var _cache = typeof Map === "function" ? new Map() : undefined;

  _wrapNativeSuper = function _wrapNativeSuper(Class) {
    if (Class === null || !_isNativeFunction(Class)) return Class;

    if (typeof Class !== "function") {
      throw new TypeError("Super expression must either be null or a function");
    }

    if (typeof _cache !== "undefined") {
      if (_cache.has(Class)) return _cache.get(Class);

      _cache.set(Class, Wrapper);
    }

    function Wrapper() {
      return _construct(Class, arguments, _getPrototypeOf(this).constructor);
    }

    Wrapper.prototype = Object.create(Class.prototype, {
      constructor: {
        value: Wrapper,
        enumerable: false,
        writable: true,
        configurable: true
      }
    });
    return _setPrototypeOf(Wrapper, Class);
  };

  return _wrapNativeSuper(Class);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

var lines = function lines(strings, joiner) {
  if (joiner === void 0) {
    joiner = '\n';
  }

  var existingLines = strings.filter(Boolean);

  if (existingLines.length) {
    return existingLines.join(joiner);
  }

  return '';
};
var indent = function indent(line) {
  return line && line.split('\n').map(function (l) {
    return "  " + l;
  }).join('\n');
};

/**
 * Statement builders for AQL
 */

var buildLimit = function buildLimit(count, skip) {
  if (skip) {
    return "LIMIT " + skip + " " + count;
  }

  return "LIMIT " + count;
};
var buildFilter = function buildFilter(condition) {
  return "FILTER " + condition;
};
var buildSort = function buildSort(property, order, sortOn) {
  if (order === void 0) {
    order = 'ASC';
  }

  if (sortOn === void 0) {
    sortOn = '$field';
  }

  return "SORT " + sortOn + "[" + interpolationOrString(property) + "] " + order;
};
var buildQueryModifiers = function buildQueryModifiers(_ref) {
  var limit = _ref.limit,
      filter = _ref.filter,
      sort = _ref.sort;
  return lines([filter && buildFilter(filter), sort && buildSort(sort.property, sort.order, sort.sortOn), limit && buildLimit(limit.count, limit.skip)]);
};
var buildSubquery = function buildSubquery(contents, returnsList) {
  return lines([(returnsList ? '' : 'FIRST') + "(", indent(contents), ")"]);
};
var buildOptions = function buildOptions(options) {
  if (!options) {
    return '';
  }

  var pairs = [options.bfs !== undefined ? "bfs: " + options.bfs : '', options.uniqueVertices !== undefined ? "uniqueVertices: \"" + options.uniqueVertices + "\"" : '', options.uniqueEdges !== undefined ? "uniqueEdges: \"" + options.uniqueEdges + "\"" : ''].filter(Boolean).join(', ');
  return "OPTIONS { " + pairs + " }";
};

var interpolationOrString = function interpolationOrString(value) {
  return value.startsWith('$') ? value : JSON.stringify(value);
};

var aqlDocument = {
  name: 'aqlDocument',
  build: function build(_ref) {
    var directiveArgs = _ref.directiveArgs,
        returnsList = _ref.returnsList,
        children = _ref.children;
    var collection = directiveArgs.collection,
        key = directiveArgs.key;

    if (returnsList) {
      return buildSubquery(lines(["FOR $field IN " + collection, indent(buildQueryModifiers(directiveArgs)), children()]), returnsList);
    } // for a singular field without a key arg, we just take the first
    // item out of the list


    if (!returnsList && !key) {
      return buildSubquery(lines(["LET $field = FIRST(", indent("FOR $field_i IN " + collection), indent(buildQueryModifiers(_extends({}, directiveArgs, {
        limit: {
          count: '1'
        }
      }))), indent("RETURN $field_i"), ")", children()]), returnsList);
    } // possibly dangerous? a check to see if this is meant to be an interpolation
    // or if we need to treat it as a literal string


    var resolvedKey = key.startsWith('$') ? key : "\"" + key + "\"";
    return buildSubquery(lines(["$field = DOCUMENT(" + collection + ", " + resolvedKey + ")", children()]), returnsList);
  }
};

var aqlNode = {
  name: 'aqlNode',
  build: function build(_ref) {
    var directiveArgs = _ref.directiveArgs,
        returnsList = _ref.returnsList,
        children = _ref.children;
    var direction = directiveArgs.direction,
        edgeCollection = directiveArgs.edgeCollection,
        options = directiveArgs.options;
    return buildSubquery(lines(["FOR $field IN " + direction + " $parent " + edgeCollection, indent(buildOptions(options)), indent(buildQueryModifiers(_extends({}, directiveArgs, {
      limit: returnsList ? directiveArgs.limit : {
        count: '1'
      }
    }))), children()]), returnsList);
  }
};

var aqlEdge = {
  name: 'aqlEdge',
  build: function build(_ref) {
    var directiveArgs = _ref.directiveArgs,
        returnsList = _ref.returnsList,
        children = _ref.children;
    var direction = directiveArgs.direction,
        collection = directiveArgs.collection,
        options = directiveArgs.options;
    return buildSubquery(lines(["FOR $field_node, $field IN " + direction + " $parent " + collection, indent(buildOptions(options)), indent(buildQueryModifiers(_extends({}, directiveArgs, {
      // enforce count 1 if this only resolves a single value
      limit: returnsList ? directiveArgs.limit : {
        count: '1'
      }
    }))), children()]), returnsList);
  }
};

var aqlEdgeNode = {
  name: 'aqlEdgeNode',
  build: function build(_ref) {
    var parentName = _ref.parentName;
    // this is assuming we are in the scope of a parent @edge subquery
    return parentName + "_node";
  }
};

var aql = {
  name: 'aql',
  build: function build(_ref) {
    var directiveArgs = _ref.directiveArgs;
    return directiveArgs.expression;
  }
};

var aqlSubquery = {
  name: 'aqlSubquery',
  build: function build(_ref) {
    var directiveArgs = _ref.directiveArgs,
        returnsList = _ref.returnsList,
        children = _ref.children;
    var query = directiveArgs.query,
        ret = directiveArgs["return"];
    return buildSubquery(lines([query, // if the user uses the "return" helper arg, we construct the right
    // field binding assignment for them to be returned by the return
    // projection construction
    ret && (returnsList ? "FOR $field IN " + ret : "LET $field = " + ret), children()]), returnsList);
  }
};

var aqlKey = {
  name: 'aqlKey',
  build: function build() {
    return "$parent._key";
  }
};

var aqlRelayConnection = {
  name: 'aqlRelayConnection',
  build: function build(_ref) {
    var directiveArgs = _ref.directiveArgs,
        returnsList = _ref.returnsList,
        children = _ref.children;
    var edgeCollection = directiveArgs.edgeCollection,
        documentCollection = directiveArgs.documentCollection;

    if (!documentCollection && !edgeCollection) {
      throw new Error('Either edgeCollection or documentCollection must be supplied to a Relay collection directive');
    }

    var listPlusOneSubquery = createListPlusOneSubquery(directiveArgs);
    return buildSubquery(lines(["LET $field_listPlusOne = " + listPlusOneSubquery, "LET $field = {", indent("edges: SLICE($field_listPlusOne, 0, $args.first)"), indent("pageInfo: { "), indent(indent("hasNextPage: LENGTH($field_listPlusOne) == $args.first + 1")), indent('}'), "}", children()]), returnsList);
  }
};

var createListPlusOneSubquery = function createListPlusOneSubquery(directiveArgs) {
  var cursorProperty = directiveArgs.cursorProperty,
      edgeDirection = directiveArgs.edgeDirection,
      edgeCollection = directiveArgs.edgeCollection,
      documentCollection = directiveArgs.documentCollection,
      source = directiveArgs.source,
      fullTextTerm = directiveArgs.fullTextTerm,
      fullTextProperty = directiveArgs.fullTextProperty,
      cursorOnEdge = directiveArgs.cursorOnEdge;

  if (source === 'default' && documentCollection) {
    if (cursorOnEdge) {
      throw new Error("Cannot use cursorOnEdge when Relay connection represents a basic document collection");
    }

    return buildSubquery(lines(["FOR $field_node IN " + documentCollection, "FILTER $field_node." + cursorProperty + " > $args.after", "SORT $field_node." + cursorProperty, "LIMIT $args.first + 1", "RETURN { cursor: $field_node." + cursorProperty + ", node: $field_node }"]), true);
  }

  if (source === 'FullText') {
    if (cursorOnEdge) {
      throw new Error("Cannot use cursorOnEdge when Relay connection is using FullText source");
    }

    if (!fullTextTerm || !fullTextProperty || !documentCollection) {
      throw new Error('fullTextTerm, fullTextProperty, and documentCollection are both required for a fulltext Relay connection directive');
    }

    return buildSubquery(lines(["FOR $field_node IN FULLTEXT(" + documentCollection + ", " + JSON.stringify(fullTextProperty) + ", " + fullTextTerm + ")", "FILTER $field_node." + cursorProperty + " > $args.after", "SORT $field_node." + cursorProperty, "LIMIT $args.first + 1", "RETURN { cursor: $field_node." + cursorProperty + ", node: $field_node }"]), true);
  }

  var cursorExpression = cursorOnEdge ? "$field_edge." + cursorProperty : "$field_node." + cursorProperty;
  return buildSubquery(lines(["FOR $field_node, $field_edge IN " + edgeDirection + " $parent " + edgeCollection, indent("OPTIONS {bfs: true}"), indent("PRUNE " + cursorExpression + " > $args.after"), "SORT " + cursorExpression, "LIMIT $args.first + 1", "RETURN MERGE($field_edge, { cursor: " + cursorExpression + ", node: $field_node })"]), true);
};

var aqlRelayEdges = {
  name: 'aqlRelayEdges',
  build: function build(_ref) {
    var returnsList = _ref.returnsList,
        children = _ref.children;
    return buildSubquery(lines(["FOR $field IN $parent.edges", children()]), returnsList);
  }
};

var aqlRelayPageInfo = {
  name: 'aqlRelayPageInfo',
  build: function build(_ref) {
    var returnsList = _ref.returnsList,
        children = _ref.children;
    return buildSubquery(lines(["LET $field = $parent.pageInfo", children()]), returnsList);
  }
};

var aqlRelayNode = {
  name: 'aqlRelayNode',
  build: function build(_ref) {
    var returnsList = _ref.returnsList,
        children = _ref.children;
    return buildSubquery(lines(["LET $field = $parent.node", children()]), returnsList);
  }
};

var defaultPlugins = {
  aqlDocument: aqlDocument,
  aqlNode: aqlNode,
  aqlEdge: aqlEdge,
  aqlEdgeNode: aqlEdgeNode,
  aql: aql,
  aqlSubquery: aqlSubquery,
  aqlKey: aqlKey,
  aqlRelayConnection: aqlRelayConnection,
  aqlRelayEdges: aqlRelayEdges,
  aqlRelayPageInfo: aqlRelayPageInfo,
  aqlRelayNode: aqlRelayNode
};

var IGNORED_FIELD_NAMES = ['__typename'];
var FIELD_PARAM_PREFIX = 'field_';

var FieldMissingError =
/*#__PURE__*/
function (_Error) {
  _inheritsLoose(FieldMissingError, _Error);

  function FieldMissingError(typeName, fieldName) {
    return _Error.call(this, "Invalid state: field \"" + fieldName + "\" does not exist on type \"" + typeName + "\"") || this;
  }

  return FieldMissingError;
}(_wrapNativeSuper(Error));

var getArgumentsPlusDefaults = function getArgumentsPlusDefaults(parentTypeName, field, schema, variables) {
  var schemaType = schema.getType(parentTypeName);

  if (!schemaType || !isObjectType(schemaType)) {
    throw new Error("Invalid state: Unknown or non-object type name \"" + parentTypeName + "\" (type: " + schemaType + ")");
  }

  var schemaField = schemaType.getFields()[field.name.value];

  if (!schemaField) {
    throw new FieldMissingError(schemaType.name, field.name.value);
  }

  var defaults = schemaField.args.reduce(function (argMap, arg) {
    var _extends2;

    return arg.defaultValue !== undefined ? _extends({}, argMap, (_extends2 = {}, _extends2[arg.name] = arg.defaultValue, _extends2)) : argMap;
  }, {});
  return _extends({}, defaults, {}, argFieldsToValues({}, field.arguments || [], variables));
};
var argFieldsToValues = function argFieldsToValues(providedValues, fields, variables) {
  return fields.reduce(function (acc, fieldNode) {
    acc[fieldNode.name.value] = valueNodeToValue(fieldNode.value, variables);
    return acc;
  }, providedValues);
};
var valueNodeToValue = function valueNodeToValue(valueNode, variables) {
  if (valueNode.kind === 'Variable') {
    return variables[valueNode.name.value];
  } else if (valueNode.kind === 'NullValue') {
    return null;
  } else if (valueNode.kind === 'ObjectValue') {
    return argFieldsToValues({}, valueNode.fields, variables);
  } else if (valueNode.kind === 'ListValue') {
    return valueNode.values.map(function (value) {
      return valueNodeToValue(value, variables);
    });
  } else if (valueNode.kind === 'IntValue') {
    return parseInt(valueNode.value, 10);
  } else if (valueNode.kind === 'FloatValue') {
    return parseFloat(valueNode.value);
  } else {
    return valueNode.value;
  }
};
var isListOrWrappedListType = function isListOrWrappedListType(type) {
  if (isListType(type)) {
    return true;
  }

  if (isNonNullType(type)) {
    return isListOrWrappedListType(type.ofType);
  }

  return false;
};
var getNameOrAlias = function getNameOrAlias(field) {
  return field.alias ? field.alias.value : field.name.value;
};
var extractObjectType = function extractObjectType(type) {
  if (isObjectType(type)) {
    return type;
  } // TODO: interface / union ?


  if (isNonNullType(type) || isListType(type)) {
    return extractObjectType(type.ofType);
  }

  return null;
};
/**
 * Converts a path from `info` into a field path, skipping over
 * array indices (since they are not represented in the schema
 * field selection paths)
 */

function getFieldPath(info) {
  var path = [];
  var pathLink = info.path;

  while (pathLink) {
    if (typeof pathLink.key === 'string') {
      path.unshift(pathLink.key);
    }

    pathLink = pathLink.prev;
  }

  return path;
}

var getDirectiveArgs = function getDirectiveArgs(directive, variableValues) {
  if (!directive || !directive.arguments) {
    return {};
  }

  return directive.arguments.reduce(function (args, arg) {
    var _extends2;

    return _extends({}, args, (_extends2 = {}, _extends2[arg.name.value] = valueNodeToValue(arg.value, variableValues), _extends2));
  }, {});
};
var getFieldDirectives = function getFieldDirectives(parentType, fieldName) {
  if (!isObjectType(parentType)) {
    throw new Error("Cannot traverse field \"" + fieldName + "\" on non-object type \"" + parentType + "\"");
  }

  var field = parentType.getFields()[fieldName];

  if (!field) {
    throw new Error("Field \"" + fieldName + "\" does not exist on type \"" + parentType.name + "\"");
  }

  if (!field.astNode) {
    throw new Error("Field \"" + parentType.name + "." + field + "\" doesn't have an AST node");
  }

  return field.astNode.directives || [];
};

var extractQueriesFromResolveInfo = function extractQueriesFromResolveInfo(_ref) {
  var info = _ref.info,
      _ref$plugins = _ref.plugins,
      plugins = _ref$plugins === void 0 ? defaultPlugins : _ref$plugins,
      _ref$argumentResolver = _ref.argumentResolvers,
      argumentResolvers = _ref$argumentResolver === void 0 ? {} : _ref$argumentResolver;
  return extractQueriesFromField({
    info: info,
    parentQuery: undefined,
    parentType: info.parentType,
    field: info.fieldNodes[0],
    path: getFieldPath(info),
    plugins: plugins,
    argumentResolvers: argumentResolvers
  });
};
var extractQueriesFromField = function extractQueriesFromField(_ref2) {
  var info = _ref2.info,
      parentQuery = _ref2.parentQuery,
      parentType = _ref2.parentType,
      field = _ref2.field,
      path = _ref2.path,
      plugins = _ref2.plugins,
      argumentResolvers = _ref2.argumentResolvers;
  var fieldName = field.name.value;

  if (IGNORED_FIELD_NAMES.includes(fieldName)) {
    return null;
  }

  if (parentQuery) {
    parentQuery.fieldNames.push(fieldName);
  }

  var schemaFieldDef = getFieldDef(info.schema, parentType, fieldName);

  if (!schemaFieldDef) {
    throw new Error("Invalid state: there's no field definition for field \"" + fieldName + "\" on type \"" + parentType.name + "\"");
  }

  var directives = getFieldDirectives(parentType, fieldName);
  var pluginInstances = directives.map(function (directive) {
    var matchingPlugin = plugins[directive.name.value];

    if (!matchingPlugin) {
      return null;
    }

    return {
      plugin: matchingPlugin,
      directiveArgs: getDirectiveArgs(directive, info.variableValues)
    };
  }).filter(Boolean);

  if (!pluginInstances.length) {
    return null;
  }

  var argValues = getArgumentsPlusDefaults(parentType.name, field, info.schema, info.variableValues);
  var paramNames = [];
  var params = {};

  if (Object.keys(argValues).length) {
    paramNames.push('args'); // process via arg resolver if it exists

    var argResolver = pathOr(function (a) {
      return a;
    }, path, argumentResolvers);
    var resolvedArgs = argResolver(argValues);
    params.args = resolvedArgs;
  }

  var baseQuery = {
    returnsList: isListOrWrappedListType(schemaFieldDef.type),
    plugins: pluginInstances,
    paramNames: paramNames,
    params: params,
    fieldNames: [],
    fieldQueries: {}
  };

  if (!field.selectionSet) {
    return baseQuery;
  }

  var currentTypeAsObjectType = extractObjectType(schemaFieldDef.type);

  if (!currentTypeAsObjectType) {
    return baseQuery;
  }

  baseQuery.fieldQueries = extractQueriesFromSelectionSet({
    selectionSet: field.selectionSet,
    parentQuery: baseQuery,
    parentType: currentTypeAsObjectType,
    info: info,
    path: path,
    plugins: plugins,
    argumentResolvers: argumentResolvers
  });
  return baseQuery;
};
var extractQueriesFromSelectionSet = function extractQueriesFromSelectionSet(_ref3) {
  var selectionSet = _ref3.selectionSet,
      path = _ref3.path,
      rest = _objectWithoutPropertiesLoose(_ref3, ["selectionSet", "path"]);

  return selectionSet.selections.reduce(function (reducedQueries, selection) {
    if (selection.kind === 'Field') {
      var _extends2;

      return _extends({}, reducedQueries, (_extends2 = {}, _extends2[getNameOrAlias(selection)] = extractQueriesFromField(_extends({
        field: selection,
        path: [].concat(path, [getNameOrAlias(selection)])
      }, rest)), _extends2));
    } else if (selection.kind === 'InlineFragment') {
      return _extends({}, reducedQueries, {}, extractQueriesFromSelectionSet(_extends({
        selectionSet: selection.selectionSet,
        path: path
      }, rest)));
    } else {
      var fragment = rest.info.fragments[selection.name.value];
      return _extends({}, reducedQueries, {}, extractQueriesFromSelectionSet(_extends({
        selectionSet: fragment.selectionSet,
        path: path
      }, rest)));
    }
  }, {});
};

// A type of promise-like that resolves synchronously and supports only one observer

const _iteratorSymbol = /*#__PURE__*/ typeof Symbol !== "undefined" ? (Symbol.iterator || (Symbol.iterator = Symbol("Symbol.iterator"))) : "@@iterator";

const _asyncIteratorSymbol = /*#__PURE__*/ typeof Symbol !== "undefined" ? (Symbol.asyncIterator || (Symbol.asyncIterator = Symbol("Symbol.asyncIterator"))) : "@@asyncIterator";

// Asynchronously call a function and send errors to recovery continuation
function _catch(body, recover) {
	try {
		var result = body();
	} catch(e) {
		return recover(e);
	}
	if (result && result.then) {
		return result.then(void 0, recover);
	}
	return result;
}

var createFieldArgGetter = function createFieldArgGetter(fieldName) {
  return function (argPath) {
    return argPath.replace('$args.', "@" + FIELD_PARAM_PREFIX + fieldName + ".args.");
  };
};
/**
 * Creates a function which replaces all "$arg.foo.bar" with the real
 * argument string
 */

var createArgReplacer = function createArgReplacer(argGetter) {
  return function (str) {
    var argMatcher = /\$args(\.\w[\w\d]+)*/;
    var result;
    var modifiedStr = '' + str;

    while ((result = argMatcher.exec(modifiedStr)) !== null) {
      var text = result[0];
      var index = result.index;
      modifiedStr = spliceString(modifiedStr, index, text, argGetter(text));
    }

    return modifiedStr;
  };
};
/**
 * Creates a function which replaces all "$field" with the actual field name
 */

var createFieldReplacer = function createFieldReplacer(fieldName) {
  return function (text) {
    return replaceAll(text, '$field', fieldName);
  };
};

var createParentReplacer = function createParentReplacer(parentName) {
  return function (text) {
    return replaceAll(text, '$parent', parentName);
  };
};

var createContextReplacer = function createContextReplacer() {
  return function (text) {
    return replaceAll(text, '$context', '@context');
  };
};

var replaceAll = function replaceAll(text, original, replacement) {
  var modifiedText = '' + text;
  var index;

  while ((index = modifiedText.indexOf(original)) >= 0) {
    modifiedText = spliceString(modifiedText, index, original, replacement);
  }

  return modifiedText;
};

var spliceString = function spliceString(text, index, original, replacement) {
  return text.slice(0, index) + replacement + text.slice(index + original.length);
};

var createAllReplacer = function createAllReplacer(_ref) {
  var fieldName = _ref.fieldName,
      parentName = _ref.parentName;
  var argReplacer = createArgReplacer(createFieldArgGetter(fieldName));
  var fieldReplacer = createFieldReplacer(fieldName);
  var parentReplacer = createParentReplacer(parentName);
  var contextReplacer = createContextReplacer();
  return function (text) {
    return contextReplacer(parentReplacer(fieldReplacer(argReplacer(text))));
  };
};

var buildQuery = function buildQuery(_ref) {
  var query = _ref.query,
      fieldName = _ref.fieldName;
  return lines(["LET query = " + buildSubQuery({
    query: query,
    fieldName: fieldName,
    parentName: '@parent'
  }), "RETURN query"]);
};
var buildSubQuery = function buildSubQuery(_ref2) {
  var query = _ref2.query,
      fieldName = _ref2.fieldName,
      parentName = _ref2.parentName;
  var statements = query.plugins.map(function (_ref3) {
    var directiveArgs = _ref3.directiveArgs,
        plugin = _ref3.plugin;
    var fieldArgs = query.params.args || {};
    var interpolate = createAllReplacer({
      fieldName: fieldName,
      parentName: parentName
    });

    var children = function children() {
      return buildReturnProjection({
        query: query,
        fieldName: fieldName
      });
    };

    return interpolate(plugin.build({
      fieldName: fieldName,
      parentName: parentName,
      fieldArgs: fieldArgs,
      directiveArgs: directiveArgs,
      returnsList: query.returnsList,
      children: children
    }));
  });
  return lines(statements);
};

var buildReturnProjection = function buildReturnProjection(_ref4) {
  var query = _ref4.query,
      fieldName = _ref4.fieldName;

  if (!query.fieldNames.length) {
    return "RETURN " + fieldName;
  }

  var scalarFields = query.fieldNames.filter(function (name) {
    return !query.fieldQueries[name];
  });
  var nonScalarFields = query.fieldNames.filter(function (name) {
    return query.fieldQueries[name];
  });
  return lines(["RETURN {", lines(scalarFields.map(function (name) {
    return name + ": " + fieldName + "." + name;
  }).map(indent), ',\n'), lines(nonScalarFields.map(function (name) {
    var fieldQuery = query.fieldQueries[name];
    var subQueryString = buildSubQuery({
      query: fieldQuery,
      fieldName: joinFieldNames(fieldName, name),
      parentName: fieldName
    });
    return name + ": " + subQueryString;
  }).map(indent), ',\n'), "}"]);
};

var joinFieldNames = function joinFieldNames(baseName, name) {
  return baseName + "_" + name;
};

/**
 * recursively flattens and builds a set of arg object variables for a query
 * and all its sub-queries
 */

var buildPrefixedFieldArgVariables = function buildPrefixedFieldArgVariables(_ref) {
  var _extends2;

  var fieldName = _ref.fieldName,
      query = _ref.query;
  return _extends((_extends2 = {}, _extends2["" + FIELD_PARAM_PREFIX + fieldName] = {
    args: query.params.args
  }, _extends2), query.fieldNames.filter(function (childFieldName) {
    return !!query.fieldQueries[childFieldName];
  }).reduce(function (args, childFieldName) {
    return _extends({}, args, {}, buildPrefixedFieldArgVariables({
      fieldName: fieldName + '_' + childFieldName,
      query: query.fieldQueries[childFieldName]
    }));
  }, {}));
};

var buildPrefixedVariables = function buildPrefixedVariables(_ref2) {
  var fieldName = _ref2.fieldName,
      query = _ref2.query,
      parent = _ref2.parent,
      contextValues = _ref2.contextValues,
      queryString = _ref2.queryString;
  return filterUnused(_extends({
    // passing the parent as a variable lets us cross the graphql -> graphdb boundary
    // and give queries access to their parent objects from our GraphQL context
    parent: parent,
    // the user may supply values in their context which they always want passed to queries
    context: contextValues
  }, buildPrefixedFieldArgVariables({
    fieldName: fieldName,
    query: query
  })), queryString);
};

var filterUnused = function filterUnused(vars, queryString) {
  return Object.keys(vars).reduce(function (filtered, key) {
    var _extends3;

    if (!new RegExp("@" + key + "\\W").test(queryString)) {
      return filtered;
    }

    return _extends({}, filtered, (_extends3 = {}, _extends3[key] = vars[key], _extends3));
  }, {});
};

var executeQuery = function executeQuery(_ref) {
  var query = _ref.query,
      bindVars = _ref.bindVars,
      db = _ref.db,
      fieldName = _ref.fieldName;

  try {
    return Promise.resolve(db.query({
      query: query,
      bindVars: bindVars
    })).then(function (queryResult) {
      return Promise.resolve(queryResult.all()).then(function (allResults) {
        return allResults[0];
      });
    });
  } catch (e) {
    return Promise.reject(e);
  }
};

var tag = '[graphql-arangodb] ';
var log = function log(_ref) {
  var title = _ref.title,
      level = _ref.level,
      details = _ref.details;

  if (level === 'verbose' && process.env.NODE_ENV !== "production" && process.env.DEBUG) {
    console.debug([chalk.yellow(tag + title)].concat(details.map(function (str) {
      return chalk.gray(str);
    })).join('\n'));
  } else if (level === 'debug' && process.env.DEBUG) {
    console.debug([chalk.cyan(tag + title)].concat(details.map(function (str) {
      return chalk.gray(str);
    })).join('\n'));
  } else if (level === 'error') {
    console.error([tag + title].concat(details).join('\n'));
  } else if (level === 'info') {
    console.info([chalk.blue(tag + title)].concat(details.map(function (str) {
      return chalk.gray(str);
    })).join('\n'));
  }
};

var runQuery = function runQuery(_ref) {
  var options = _ref.options,
      context = _ref.context,
      query = _ref.query,
      info = _ref.info,
      additionalBindVars = _ref.additionalBindVars,
      parent = _ref.parent;

  try {
    var db = options.db,
        _options$contextKey = options.contextKey,
        contextKey = _options$contextKey === void 0 ? 'arangoContext' : _options$contextKey,
        _options$contextDbKey = options.contextDbKey,
        contextDbKey = _options$contextDbKey === void 0 ? 'arangoDb' : _options$contextDbKey;
    var resolvedDb = db || context[contextDbKey];

    if (!resolvedDb) {
      throw new Error("Either a valid ArangoDB Database instance must be supplied on the \"" + contextDbKey + "\" property of the context, or you must create your own resolver using createResolver from graphql-arangodb");
    }

    return Promise.resolve(_catch(function () {
      var queryString = buildQuery({
        query: query,
        fieldName: info.fieldName
      });
      var bindVars = buildPrefixedVariables({
        fieldName: info.fieldName,
        query: query,
        parent: parent,
        contextValues: context[contextKey],
        queryString: queryString
      });
      log({
        title: "Running query",
        level: 'info',
        details: [queryString, JSON.stringify(bindVars)]
      });
      return Promise.resolve(executeQuery({
        query: queryString,
        bindVars: _extends({}, additionalBindVars || {}, {}, bindVars),
        db: resolvedDb,
        fieldName: info.fieldName
      })).then(function (data) {
        log({
          title: "Query response data",
          level: 'debug',
          details: [JSON.stringify(data)]
        });
        return data;
      });
    }, function (err) {
      log({
        title: "Query execution error",
        level: 'error',
        details: [err.toString(), err.stack]
      });
      throw err;
    }));
  } catch (e) {
    return Promise.reject(e);
  }
};

var createCustomQueryRunner = function createCustomQueryRunner(options) {
  return function (_ref) {
    var queryString = _ref.queryString,
        info = _ref.info,
        context = _ref.context,
        parent = _ref.parent,
        bindVars = _ref.bindVars;

    try {
      var _options$plugins = options.plugins,
          plugins = _options$plugins === void 0 ? defaultPlugins : _options$plugins,
          _options$argumentReso = options.argumentResolvers,
          argumentResolvers = _options$argumentReso === void 0 ? {} : _options$argumentReso;
      var customQueryPlugin = {
        name: 'customQuery',
        build: function build(_ref2) {
          var children = _ref2.children,
              returnsList = _ref2.returnsList;
          return buildSubquery(lines(["LET $field = " + buildSubquery(queryString, returnsList), children()]), returnsList);
        }
      };
      var query = {
        returnsList: isListOrWrappedListType(info.returnType),
        plugins: [{
          plugin: customQueryPlugin,
          directiveArgs: {}
        }],
        paramNames: [],
        params: {},
        fieldNames: [],
        fieldQueries: {}
      };
      var selectionSet = info.fieldNodes[0].selectionSet;
      var returnTypeAsObjectType = extractObjectType(info.returnType);

      if (!selectionSet || !returnTypeAsObjectType) {
        throw new Error('Not implemented: custom query without GraphQL selection or Object return type');
      }

      query.fieldQueries = extractQueriesFromSelectionSet({
        selectionSet: selectionSet,
        info: info,
        path: getFieldPath(info),
        parentQuery: query,
        parentType: returnTypeAsObjectType,
        plugins: plugins,
        argumentResolvers: argumentResolvers
      });
      return Promise.resolve(runQuery({
        options: options,
        context: context,
        query: query,
        info: info,
        parent: parent,
        additionalBindVars: bindVars
      }));
    } catch (e) {
      return Promise.reject(e);
    }
  };
};
var runCustomQuery =
/*#__PURE__*/
createCustomQueryRunner({});

var createResolver = function createResolver(options) {
  var aqlResolver = function aqlResolver(parent, args, context, info) {
    try {
      var _options$plugins = options.plugins,
          plugins = _options$plugins === void 0 ? defaultPlugins : _options$plugins,
          _options$argumentReso = options.argumentResolvers,
          argumentResolvers = _options$argumentReso === void 0 ? {} : _options$argumentReso;
      var query = extractQueriesFromResolveInfo({
        info: info,
        plugins: plugins,
        argumentResolvers: argumentResolvers
      });

      if (!query) {
        return Promise.resolve(null);
      }

      return Promise.resolve(runQuery({
        options: options,
        info: info,
        query: query,
        context: context,
        parent: parent
      }));
    } catch (e) {
      return Promise.reject(e);
    }
  };
  /**
   * Construct your own AQL query within a resolver, using whatever logic you wish,
   * then pass it to this function (along with bindVars). Also pass in the parent,
   * context and info arguments of your resolver. This function will take your original
   * query and append additional AQL to it so that it will resolve the rest of the
   * GraphQL query selections the user has made.
   */


  aqlResolver.runCustomQuery = createCustomQueryRunner(options);
  return aqlResolver;
};
var resolver =
/*#__PURE__*/
createResolver({});

var directiveTypeDefs = "\n\"\"\"Represents the direction of an edge in the graph relative to the current node\"\"\"\nenum AqlEdgeDirection {\n  OUTBOUND\n  INBOUND\n  ANY\n}\n\n\"\"\"Represents the order of a sorting operation\"\"\"\nenum AqlSortOrder {\n  DESC\n  ASC\n}\n\nenum AqlRelayConnectionSource {\n  Default\n  FullText\n}\n\ninput AqlSortInput {\n  \"\"\"The property to sort on\"\"\"\n  property: String!\n  \"\"\"The order to sort in. Defaults ASC\"\"\"\n  order: AqlSortOrder = ASC\n  \"\"\"Change the object being sorted. Defaults to $field\"\"\"\n  sortOn: String\n}\n\ninput AqlLimitInput {\n  \"\"\"The upper limit of documents to return\"\"\"\n  count: String!\n  \"\"\"The number of documents to skip\"\"\"\n  skip: String\n}\n\ninput AqlTraverseOptionsInput {\n  \"\"\"Enables breadth-first search\"\"\"\n  bfs: Boolean\n  \"\"\"\n  - \"path\": guarantees no path is returned with a duplicate vertex\n  - \"global\": guarantees each vertex is visited at most once for the whole traversal\n  - \"none\": (default) no uniqueness check\n  \"\"\"\n  uniqueVertices: String\n  \"\"\"\n  - \"path\": (default) guarantees no path is returned with a duplicate edge\n  - \"none\": allows paths to 'double back' onto edges cyclically\n  \"\"\"\n  uniqueEdges: String\n}\n\ndirective @aqlDocument(\n  collection: String!\n  key: String\n  filter: String\n  sort: AqlSortInput\n  limit: AqlLimitInput\n) on FIELD_DEFINITION\n\ndirective @aqlNode(\n  edgeCollection: String!\n  direction: AqlEdgeDirection!\n  filter: String\n  sort: AqlSortInput\n  limit: AqlLimitInput\n  options: AqlTraverseOptionsInput\n) on FIELD_DEFINITION\n\ndirective @aqlEdge(\n  direction: AqlEdgeDirection!\n  collection: String!\n  filter: String\n  sort: AqlSortInput\n  limit: AqlLimitInput\n  options: AqlTraverseOptionsInput\n) on FIELD_DEFINITION\n\ndirective @aqlEdgeNode on FIELD_DEFINITION\n\ndirective @aql(\n  expression: String!\n) on FIELD_DEFINITION\n\ndirective @aqlSubquery(\n  query: String!\n  return: String\n) on FIELD_DEFINITION\n\ndirective @aqlKey on FIELD_DEFINITION\n\ndirective @aqlRelayConnection(\n  edgeCollection: String\n  edgeDirection: AqlEdgeDirection\n  documentCollection: String\n  cursorProperty: String!\n  source: AqlRelayConnectionSource = Default\n  fullTextTerm: String\n  fullTextProperty: String\n  cursorOnEdge: Boolean = false\n) on FIELD_DEFINITION | OBJECT\n\ndirective @aqlRelayEdges on FIELD_DEFINITION | OBJECT\n\ndirective @aqlRelayPageInfo on FIELD_DEFINITION | OBJECT\n\ndirective @aqlRelayNode on FIELD_DEFINITION | OBJECT\n";

export default resolver;
export { createCustomQueryRunner, createResolver, directiveTypeDefs, defaultPlugins as plugins, resolver, runCustomQuery };
//# sourceMappingURL=graphql-arangodb.esm.js.map
